@@ -8,8 +8,10 @@
 #include <cairo/cairo.h>
 #include <pango/pangocairo.h>
 #include <stdarg.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <math.h>
 
 #ifdef IMV_BACKEND_LIBRSVG
 #include <librsvg/rsvg.h>
@@ -203,7 +205,9 @@ static int convert_pixelformat(enum imv_pixelformat fmt)
 static void draw_bitmap(struct imv_canvas *canvas,
                         struct imv_bitmap *bitmap,
                         int bx, int by, double scale,
-                        enum upscaling_method upscaling_method)
+                        double rotation, bool mirrored,
+                        enum upscaling_method upscaling_method,
+                        bool cache_invalidated)
 {
   GLint viewport[4];
   glGetIntegerv(GL_VIEWPORT, viewport);
@@ -219,9 +223,19 @@ static void draw_bitmap(struct imv_canvas *canvas,
 
   glBindTexture(GL_TEXTURE_RECTANGLE, canvas->cache.texture);
 
-  if (canvas->cache.bitmap != bitmap) {
-    glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
-    glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
+  GLint upscaling = 0;
+  if (upscaling_method == UPSCALING_LINEAR) {
+    upscaling = GL_LINEAR;
+  } else if (upscaling_method == UPSCALING_NEAREST_NEIGHBOUR) {
+    upscaling = GL_NEAREST;
+  } else {
+    imv_log(IMV_ERROR, "Unknown upscaling method: %d\n", upscaling_method);
+    abort();
+  }
+
+  if (canvas->cache.bitmap != bitmap || cache_invalidated) {
+    glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, upscaling);
+    glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, upscaling);
     glPixelStorei(GL_UNPACK_ROW_LENGTH, bitmap->width);
     glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
     glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
@@ -232,19 +246,21 @@ static void draw_bitmap(struct imv_canvas *canvas,
 
   glEnable(GL_TEXTURE_RECTANGLE);
 
-  if (upscaling_method == UPSCALING_LINEAR) {
-    glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-  } else if (upscaling_method == UPSCALING_NEAREST_NEIGHBOUR) {
-    glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
-  } else {
-    imv_log(IMV_ERROR, "Unknown upscaling method: %d\n", upscaling_method);
-    abort();
-  }
+  glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, upscaling);
 
   const int left = bx;
   const int top = by;
   const int right = left + bitmap->width * scale;
   const int bottom = top + bitmap->height * scale;
+  const int center_x = left + bitmap->width * scale / 2;
+  const int center_y = top + bitmap->height * scale / 2;
+
+  glTranslated(center_x, center_y, 0);
+  if (mirrored) {
+    glScaled(-1, 1, 1);
+  }
+  glRotated(rotation, 0, 0, 1);
+  glTranslated(-center_x, -center_y, 0);
 
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
@@ -269,11 +285,14 @@ RsvgHandle *imv_image_get_svg(const struct imv_image *image);
 
 void imv_canvas_draw_image(struct imv_canvas *canvas, struct imv_image *image,
                            int x, int y, double scale,
-                           enum upscaling_method upscaling_method)
+                           double rotation, bool mirrored,
+                           enum upscaling_method upscaling_method,
+                           bool cache_invalidated)
 {
   struct imv_bitmap *bitmap = imv_image_get_bitmap(image);
   if (bitmap) {
-    draw_bitmap(canvas, bitmap, x, y, scale, upscaling_method);
+    draw_bitmap(canvas, bitmap, x, y, scale, rotation, mirrored,
+                upscaling_method, cache_invalidated);
     return;
   }
 
@@ -283,6 +302,12 @@ void imv_canvas_draw_image(struct imv_canvas *canvas, struct imv_image *image,
     imv_canvas_clear(canvas);
     cairo_translate(canvas->cairo, x, y);
     cairo_scale(canvas->cairo, scale, scale);
+    cairo_translate(canvas->cairo, imv_image_width(image) / 2, imv_image_height(image) / 2);
+    if (mirrored) {
+      cairo_scale(canvas->cairo, -1, 1);
+    }
+    cairo_rotate(canvas->cairo, rotation * M_PI / 180.0);
+    cairo_translate(canvas->cairo, -imv_image_width(image) / 2, -imv_image_height(image) / 2);
     rsvg_handle_render_cairo(svg, canvas->cairo);
     cairo_identity_matrix(canvas->cairo);
     imv_canvas_draw(canvas);
