@@ -4,6 +4,7 @@
 #include <ctype.h>
 #include <errno.h>
 #include <getopt.h>
+#include <limits.h>
 #include <pthread.h>
 #include <stdbool.h>
 #include <stdint.h>
@@ -52,6 +53,10 @@ enum internal_event_type {
   COMMAND
 };
 
+struct color_rgb {
+  unsigned char r, g, b;
+};
+
 struct internal_event {
   enum internal_event_type type;
   union {
@@ -83,8 +88,27 @@ struct imv {
   int initial_width;
   int initial_height;
 
-  /* display some textual info onscreen */
-  bool overlay_enabled;
+  /* overlay */
+  struct {
+    /* display some textual info onscreen */
+    bool enabled;
+    /* the user-specified format strings for the overlay*/
+    char *text;
+    struct color_rgb text_color;
+    unsigned char text_alpha;
+    struct color_rgb background_color;
+    unsigned char background_alpha;
+
+    /* overlay position */
+    bool position_at_bottom;
+
+    /* overlay font */
+    struct {
+      char *name;
+      int size;
+    } font;
+  } overlay;
+
 
   /* method for scaling up images: interpolate or nearest neighbour */
   enum upscaling_method upscaling_method;
@@ -92,6 +116,7 @@ struct imv {
   /* dirty state flags */
   bool need_redraw;
   bool need_rescale;
+  bool cache_invalidated;
 
   /* traverse sub-directories for more images */
   bool recursive_load;
@@ -105,6 +130,9 @@ struct imv {
   /* read paths from stdin, as opposed to image data */
   bool paths_from_stdin;
 
+  /* FILE to use when reading paths from stdin */
+  FILE *stdin_pipe;
+
   /* scale up / down images to match window, or actual size */
   enum scaling_mode scaling_mode;
 
@@ -116,7 +144,7 @@ struct imv {
     /* show a solid background colour, or chequerboard pattern */
     enum background_type type;
     /* the aforementioned background colour */
-    struct { unsigned char r, g, b; } color;
+    struct color_rgb color;
   } background;
 
   /* slideshow state tracking */
@@ -138,11 +166,6 @@ struct imv {
 
   struct imv_image *current_image;
 
-  /* overlay font */
-  struct {
-    char *name;
-    int size;
-  } font;
 
   /* if specified by user, the path of the first image to display */
   char *starting_path;
@@ -152,7 +175,6 @@ struct imv {
 
   /* the user-specified format strings for the overlay and window title */
   char *title_text;
-  char *overlay_text;
 
   /* imv subsystems */
   struct imv_binds *binds;
@@ -178,6 +200,8 @@ static void command_next(struct list *args, const char *argstr, void *data);
 static void command_prev(struct list *args, const char *argstr, void *data);
 static void command_goto(struct list *args, const char *argstr, void *data);
 static void command_zoom(struct list *args, const char *argstr, void *data);
+static void command_rotate(struct list *args, const char *argstr, void *data);
+static void command_flip(struct list *args, const char *argstr, void *data);
 static void command_open(struct list *args, const char *argstr, void *data);
 static void command_close(struct list *args, const char *argstr, void *data);
 static void command_fullscreen(struct list *args, const char *argstr, void *data);
@@ -395,9 +419,14 @@ static void key_handler(struct imv *imv, const struct imv_event *event)
       return;
     }
 
-    struct list *cmds = imv_bind_handle_event(imv->binds, event->data.keyboard.description);
-    if (cmds) {
-      imv_command_exec_list(imv->commands, cmds, imv);
+    /* Keys such as Shift on their own come through as '', we should skip them
+     * since they'll turn up later as 'Shift+W', etc.
+     */
+    if (*event->data.keyboard.description != '\0') {
+      struct list *cmds = imv_bind_handle_event(imv->binds, event->data.keyboard.description);
+      if (cmds) {
+        imv_command_exec_list(imv->commands, cmds, imv);
+      }
     }
   }
 
@@ -448,11 +477,42 @@ static void event_handler(void *data, const struct imv_event *e)
 
 }
 
+static bool hex_value_to_color_rgb(const char* hex, struct color_rgb* color)
+{
+    if (*hex == '#')
+      hex++;
+
+    char *ep;
+    uint32_t n = strtoul(hex, &ep, 16);
+    if (*ep != '\0' || ep - hex != 6 || n > 0xFFFFFF) {
+      imv_log(IMV_ERROR, "Invalid hex color: '%s'\n", hex);
+      return false;
+    }
+    color->b = n & 0xFF;
+    color->g = (n >> 8) & 0xFF;
+    color->r = (n >> 16);
+    return true;
+}
+
+static bool hex_value_to_alpha(const char* hex, unsigned char *alpha)
+{
+    if (*hex == '#')
+      hex++;
+    char *ep;
+    uint32_t n = strtoul(hex, &ep, 16);
+    if (*ep != '\0' || ep - hex != 2 || n > 0xFF) {
+      imv_log(IMV_ERROR, "Invalid hex color: '%s'\n", hex);
+      return false;
+    }
+    *alpha = n & 0xFF;
+    return true;
+}
+
 static void log_to_stderr(enum imv_log_level level, const char *text, void *data)
 {
   (void)data;
   if (level >= IMV_INFO) {
-    fprintf(stderr, "%s", text);
+    fputs(text, stderr);
   }
 }
 
@@ -468,8 +528,8 @@ struct imv *imv_create(void)
   imv->need_rescale = true;
   imv->scaling_mode = SCALING_FULL;
   imv->loop_input = true;
-  imv->font.name = strdup("Monospace");
-  imv->font.size = 24;
+  imv->overlay.font.name = strdup("Monospace");
+  imv->overlay.font.size = 24;
   imv->binds = imv_binds_create();
   imv->navigator = imv_navigator_create();
   imv->backends = list_create();
@@ -477,17 +537,28 @@ struct imv *imv_create(void)
   imv->console = imv_console_create();
   imv_console_set_command_callback(imv->console, &command_callback, imv);
   imv->ipc = imv_ipc_create();
-  imv_ipc_set_command_callback(imv->ipc, &command_callback, imv);
+  if (imv->ipc) {
+    imv_ipc_set_command_callback(imv->ipc, &command_callback, imv);
+  }
   imv->title_text = strdup(
       "imv - [${imv_current_index}/${imv_file_count}]"
       " [${imv_width}x${imv_height}] [${imv_scale}%]"
       " $imv_current_file [$imv_scaling_mode]"
   );
-  imv->overlay_text = strdup(
+  imv->overlay.text = strdup(
       "[${imv_current_index}/${imv_file_count}]"
       " [${imv_width}x${imv_height}] [${imv_scale}%]"
       " $imv_current_file [$imv_scaling_mode]"
   );
+  imv->overlay.text_color.r = 255;
+  imv->overlay.text_color.g = 255;
+  imv->overlay.text_color.b = 255;
+  imv->overlay.text_alpha = 255;
+  imv->overlay.background_color.r = 0;
+  imv->overlay.background_color.g = 0;
+  imv->overlay.background_color.b = 0;
+  imv->overlay.background_alpha = 195;
+  imv->overlay.position_at_bottom = false;
   imv->startup_commands = list_create();
 
   imv_command_register(imv->commands, "quit", &command_quit);
@@ -496,6 +567,8 @@ struct imv *imv_create(void)
   imv_command_register(imv->commands, "prev", &command_prev);
   imv_command_register(imv->commands, "goto", &command_goto);
   imv_command_register(imv->commands, "zoom", &command_zoom);
+  imv_command_register(imv->commands, "rotate", &command_rotate);
+  imv_command_register(imv->commands, "flip", &command_flip);
   imv_command_register(imv->commands, "open", &command_open);
   imv_command_register(imv->commands, "close", &command_close);
   imv_command_register(imv->commands, "fullscreen", &command_fullscreen);
@@ -561,9 +634,9 @@ struct imv *imv_create(void)
 
 void imv_free(struct imv *imv)
 {
-  free(imv->font.name);
+  free(imv->overlay.font.name);
   free(imv->title_text);
-  free(imv->overlay_text);
+  free(imv->overlay.text);
   imv_binds_free(imv->binds);
   imv_navigator_free(imv->navigator);
   if (imv->current_source) {
@@ -605,17 +678,7 @@ static bool parse_bg(struct imv *imv, const char *bg)
     imv->background.type = BACKGROUND_CHEQUERED;
   } else {
     imv->background.type = BACKGROUND_SOLID;
-    if (*bg == '#')
-      ++bg;
-    char *ep;
-    uint32_t n = strtoul(bg, &ep, 16);
-    if (*ep != '\0' || ep - bg != 6 || n > 0xFFFFFF) {
-      imv_log(IMV_ERROR, "Invalid hex color: '%s'\n", bg);
-      return false;
-    }
-    imv->background.color.b = n & 0xFF;
-    imv->background.color.g = (n >> 8) & 0xFF;
-    imv->background.color.r = (n >> 16);
+    return hex_value_to_color_rgb(bg, &imv->background.color);
   }
   return true;
 }
@@ -679,14 +742,32 @@ static bool parse_initial_pan(struct imv *imv, const char *pan_params)
   return true;
 }
 
+static void *pipe_stdin(void *data)
+{
+  int *fd = data;
+  while (1) {
+    char buf[PIPE_BUF];
+    ssize_t err = read(STDIN_FILENO, buf, PIPE_BUF);
+    if (err > 0) {
+      /* writes up to PIPE_BUF are atomic */
+      write(*fd, buf, err);
+    } else if (err == 0 || errno != EINTR) {
+      /* break if EOF or actual read error */
+      break;
+    }
+  }
+
+  return NULL;
+}
+
 static void *load_paths_from_stdin(void *data)
 {
   struct imv *imv = data;
 
-  imv_log(IMV_INFO, "Reading paths from stdin...");
+  imv_log(IMV_INFO, "Reading paths from stdin...\n");
 
   char buf[PATH_MAX];
-  while (fgets(buf, sizeof(buf), stdin) != NULL) {
+  while (fgets(buf, sizeof(buf), imv->stdin_pipe) != NULL) {
     size_t len = strlen(buf);
     if (buf[len-1] == '\n') {
       buf[--len] = 0;
@@ -745,7 +826,7 @@ bool imv_parse_args(struct imv *imv, int argc, char **argv)
     switch(o) {
       case 'f': imv->start_fullscreen = true;                    break;
       case 'r': imv->recursive_load = true;                      break;
-      case 'd': imv->overlay_enabled = true;                     break;
+      case 'd': imv->overlay.enabled = true;                     break;
       case 'x': imv->loop_input = false;                         break;
       case 'l': imv->list_files_at_exit = true;                  break;
       case 'n': imv->starting_path = optarg;                     break;
@@ -835,10 +916,25 @@ int imv_run(struct imv *imv)
 
   /* if loading paths from stdin, kick off a thread to do that - we'll receive
    * events back via internal events */
+  int *stdin_pipe_fds = NULL;
+  pthread_t pipe_stdin_thread;
+  pthread_t load_paths_thread;
   if (imv->paths_from_stdin) {
-    pthread_t thread;
-    pthread_create(&thread, NULL, load_paths_from_stdin, imv);
-    pthread_detach(thread);
+    /* this array is allocated on the heap because it's passed to pthread_create */
+    stdin_pipe_fds = calloc(2, sizeof *stdin_pipe_fds);
+
+    if (pipe(stdin_pipe_fds)) {
+      /* if pipe creation fails, we should exit */
+      free(stdin_pipe_fds);
+      return 1;
+    }
+
+    imv->stdin_pipe = fdopen(stdin_pipe_fds[0], "re");
+
+    if (pthread_create(&load_paths_thread, NULL, load_paths_from_stdin, imv)
+        || pthread_create(&pipe_stdin_thread, NULL, pipe_stdin, stdin_pipe_fds + 1)) {
+      return 1;
+    }
   }
 
   if (imv->starting_path) {
@@ -1043,6 +1139,21 @@ int imv_run(struct imv *imv)
       puts(imv_navigator_at(imv->navigator, i));
   }
 
+  if (imv->paths_from_stdin) {
+    /* stop the thread before closing the pipe */
+    pthread_cancel(pipe_stdin_thread);
+    pthread_join(pipe_stdin_thread, NULL);
+
+    /* will cause the thread running load_paths_from_stdin() to exit */
+    close(stdin_pipe_fds[1]);
+    free(stdin_pipe_fds);
+
+    /* join the other thread to avoid a race when closing imv->stdin_pipe */
+    pthread_join(load_paths_thread, NULL);
+
+    fclose(imv->stdin_pipe);
+  }
+
   return 0;
 }
 
@@ -1073,7 +1184,7 @@ static bool setup_window(struct imv *imv)
     int ww, wh;
     imv_window_get_size(imv->window, &ww, &wh);
     imv->canvas = imv_canvas_create(ww, wh);
-    imv_canvas_font(imv->canvas, imv->font.name, imv->font.size);
+    imv_canvas_font(imv->canvas, imv->overlay.font.name, imv->overlay.font.size);
   }
 
   return true;
@@ -1182,29 +1293,48 @@ static void render_window(struct imv *imv)
   /* draw our actual image */
   if (imv->current_image) {
     int x, y;
-    double scale;
+    double scale, rotation;
+    bool mirrored;
     imv_viewport_get_offset(imv->view, &x, &y);
     imv_viewport_get_scale(imv->view, &scale);
-    imv_canvas_draw_image(imv->canvas, imv->current_image, x, y, scale, imv->upscaling_method);
+    imv_viewport_get_rotation(imv->view, &rotation);
+    imv_viewport_get_mirrored(imv->view, &mirrored);
+    imv_canvas_draw_image(imv->canvas, imv->current_image,
+                          x, y, scale, rotation, mirrored,
+                          imv->upscaling_method, imv->cache_invalidated);
   }
 
   imv_canvas_clear(imv->canvas);
 
   /* if the overlay needs to be drawn, draw that too */
-  if (imv->overlay_enabled) {
-    const int height = imv->font.size * 1.2;
-    imv_canvas_color(imv->canvas, 0, 0, 0, 0.75);
-    imv_canvas_fill_rectangle(imv->canvas, 0, 0, ww, height);
-    imv_canvas_color(imv->canvas, 1, 1, 1, 1);
+  if (imv->overlay.enabled) {
+    const int height = imv->overlay.font.size * 1.2;
+    imv_canvas_color(imv->canvas,
+        imv->overlay.background_color.r / 255.f,
+        imv->overlay.background_color.g / 255.f,
+        imv->overlay.background_color.b / 255.f,
+        imv->overlay.background_alpha / 255.f);
+    int y = 0 ;
+    const int bottom_offset = 5;
+    if (imv->overlay.position_at_bottom)
+    {
+      y = wh - height - bottom_offset;
+    }
+    imv_canvas_fill_rectangle(imv->canvas, 0, y, ww, height + bottom_offset);
+    imv_canvas_color(imv->canvas,
+        imv->overlay.text_color.r / 255.f,
+        imv->overlay.text_color.g / 255.f,
+        imv->overlay.text_color.b / 255.f,
+        imv->overlay.text_alpha / 255.f);
     char overlay_text[1024];
-    generate_env_text(imv, overlay_text, sizeof overlay_text, imv->overlay_text);
-    imv_canvas_printf(imv->canvas, 0, 0, "%s", overlay_text);
+    generate_env_text(imv, overlay_text, sizeof overlay_text, imv->overlay.text);
+    imv_canvas_printf(imv->canvas, 0, y, "%s", overlay_text);
   }
 
   /* draw command entry bar if needed */
   if (imv_console_prompt(imv->console)) {
     const int bottom_offset = 5;
-    const int height = imv->font.size * 1.2;
+    const int height = imv->overlay.font.size * 1.2;
     imv_canvas_color(imv->canvas, 0, 0, 0, 0.75);
     imv_canvas_fill_rectangle(imv->canvas, 0, wh - height - bottom_offset,
         ww, height + bottom_offset);
@@ -1229,6 +1359,7 @@ static void render_window(struct imv *imv)
 
   /* redraw complete, unset the flag */
   imv->need_redraw = false;
+  imv->cache_invalidated = false;
 }
 
 static char *get_config_path(void)
@@ -1306,7 +1437,7 @@ static int handle_ini_value(void *user, const char *section, const char *name,
     }
 
     if (!strcmp(name, "overlay")) {
-      imv->overlay_enabled = parse_bool(value);
+      imv->overlay.enabled = parse_bool(value);
       return 1;
     }
 
@@ -1351,22 +1482,55 @@ static int handle_ini_value(void *user, const char *section, const char *name,
       return 1;
     }
 
+    if (!strcmp(name, "overlay_text_color")) {
+      if (!hex_value_to_color_rgb(value, &imv->overlay.text_color)) {
+        return false;
+      }
+      return 1;
+    }
+
+    if (!strcmp(name, "overlay_text_alpha")) {
+      if (!hex_value_to_alpha(value, &imv->overlay.text_alpha)) {
+        return false;
+      }
+      return 1;
+    }
+
+    if (!strcmp(name, "overlay_background_color")) {
+      if (!hex_value_to_color_rgb(value, &imv->overlay.background_color)) {
+        return false;
+      }
+      return 1;
+    }
+
+    if (!strcmp(name, "overlay_background_alpha")) {
+      if (!hex_value_to_alpha(value, &imv->overlay.background_alpha)) {
+        return false;
+      }
+      return 1;
+    }
+
+    if (!strcmp(name, "overlay_position_bottom")) {
+      imv->overlay.position_at_bottom = parse_bool(value);
+      return 1;
+    }
+
     if (!strcmp(name, "overlay_font")) {
-      free(imv->font.name);
-      imv->font.name = strdup(value);
-      char *sep = strchr(imv->font.name, ':');
+      free(imv->overlay.font.name);
+      imv->overlay.font.name = strdup(value);
+      char *sep = strchr(imv->overlay.font.name, ':');
       if (sep) {
         *sep = 0;
-        imv->font.size = atoi(sep + 1);
+        imv->overlay.font.size = atoi(sep + 1);
       } else {
-        imv->font.size = 24;
+        imv->overlay.font.size = 24;
       }
       return 1;
     }
 
     if (!strcmp(name, "overlay_text")) {
-      free(imv->overlay_text);
-      imv->overlay_text = strdup(value);
+      free(imv->overlay.text);
+      imv->overlay.text = strdup(value);
       return 1;
     }
 
@@ -1447,6 +1611,7 @@ static void command_next(struct list *args, const char *argstr, void *data)
   }
 
   imv_navigator_select_rel(imv->navigator, index);
+  imv_viewport_reset_transform(imv->view);
 
   imv->slideshow.elapsed = 0;
 }
@@ -1462,6 +1627,7 @@ static void command_prev(struct list *args, const char *argstr, void *data)
   }
 
   imv_navigator_select_rel(imv->navigator, -index);
+  imv_viewport_reset_transform(imv->view);
 
   imv->slideshow.elapsed = 0;
 }
@@ -1475,7 +1641,8 @@ static void command_goto(struct list *args, const char *argstr, void *data)
   }
 
   long int index = strtol(args->items[1], NULL, 10);
-  imv_navigator_select_abs(imv->navigator, index - 1);
+  imv_navigator_select_abs(imv->navigator, index > 0 ? index - 1 : index);
+  imv_viewport_reset_transform(imv->view);
 
   imv->slideshow.elapsed = 0;
 }
@@ -1495,6 +1662,34 @@ static void command_zoom(struct list *args, const char *argstr, void *data)
   }
 }
 
+static void command_rotate(struct list *args, const char *argstr, void *data)
+{
+  (void)argstr;
+  struct imv *imv = data;
+  if (args->len == 3) {
+    if (!strcmp(args->items[1], "by")) {
+      double degrees = strtod(args->items[2], NULL);
+      imv_viewport_rotate_by(imv->view, degrees);
+    } else if (!strcmp(args->items[1], "to")) {
+      double degrees = strtod(args->items[2], NULL);
+      imv_viewport_rotate_to(imv->view, degrees);
+    }
+  }
+}
+
+static void command_flip(struct list *args, const char *argstr, void *data)
+{
+  (void)argstr;
+  struct imv *imv = data;
+  if (args->len == 2) {
+    if (!strcmp(args->items[1], "vertical")) {
+      imv_viewport_flip_v(imv->view);
+    } else if (!strcmp(args->items[1], "horizontal")) {
+      imv_viewport_flip_h(imv->view);
+    }
+  }
+}
+
 static void command_open(struct list *args, const char *argstr, void *data)
 {
   (void)argstr;
@@ -1557,7 +1752,7 @@ static void command_overlay(struct list *args, const char *argstr, void *data)
   (void)args;
   (void)argstr;
   struct imv *imv = data;
-  imv->overlay_enabled = !imv->overlay_enabled;
+  imv->overlay.enabled = !imv->overlay.enabled;
   imv->need_redraw = true;
 }
 
@@ -1582,6 +1777,7 @@ static void command_reset(struct list *args, const char *argstr, void *data)
   (void)args;
   (void)argstr;
   struct imv *imv = data;
+  imv_viewport_reset_transform(imv->view);
   imv->need_rescale = true;
   imv->need_redraw = true;
 }
@@ -1650,6 +1846,7 @@ static void command_set_upscaling_method(struct list *args, const char *argstr,
   }
 
   imv->need_redraw = true;
+  imv->cache_invalidated = true;
 }
 
 static void command_set_slideshow_duration(struct list *args, const char *argstr, void *data)
