@@ -8,6 +8,7 @@ extern const struct imv_backend imv_backend_librsvg;
 extern const struct imv_backend imv_backend_libtiff;
 extern const struct imv_backend imv_backend_libjpeg;
 extern const struct imv_backend imv_backend_libnsgif;
+extern const struct imv_backend imv_backend_libheif;
 
 int main(int argc, char **argv)
 {
@@ -41,6 +42,10 @@ int main(int argc, char **argv)
   imv_install_backend(imv, &imv_backend_freeimage);
 #endif
 
+#ifdef IMV_BACKEND_LIBHEIF
+  imv_install_backend(imv, &imv_backend_libheif);
+#endif
+
   if (!imv_load_config(imv)) {
     imv_free(imv);
     return 1;
