@@ -34,6 +34,8 @@ struct imv_window {
   EGLSurface           egl_surface;
   struct wl_egl_window *egl_window;
 
+  bool xdg_configured;
+
   struct imv_keyboard *keyboard;
   struct list *wl_outputs;
 
@@ -103,9 +105,12 @@ static void keyboard_keymap(void *data, struct wl_keyboard *keyboard,
   (void)keyboard;
   (void)format;
   struct imv_window *window = data;
-  char *src = mmap(NULL, size, PROT_READ, MAP_SHARED, fd, 0);
-  imv_keyboard_set_keymap(window->keyboard, src);
-  munmap(src, size);
+  char *src = mmap(NULL, size, PROT_READ, MAP_PRIVATE, fd, 0);
+
+  if (src != MAP_FAILED) {
+    imv_keyboard_set_keymap(window->keyboard, src);
+    munmap(src, size);
+  }
   close(fd);
 }
 
@@ -131,6 +136,7 @@ static void keyboard_leave(void *data, struct wl_keyboard *keyboard,
 static void cleanup_event(struct imv_event *event)
 {
   if (event->type == IMV_EVENT_KEYBOARD) {
+    free(event->data.keyboard.description);
     free(event->data.keyboard.keyname);
     free(event->data.keyboard.text);
   }
@@ -522,6 +528,9 @@ static void on_remove_global(void *data, struct wl_registry *registry, uint32_t
 
 static void update_scale(struct imv_window *window)
 {
+  if (!window->xdg_configured) {
+    return;
+  }
   int new_scale = 1;
   for (size_t i = 0; i < window->wl_outputs->len; ++i) {
     struct output_data *data = window->wl_outputs->items[i];
@@ -586,6 +595,8 @@ static void surface_configure(void *data, struct xdg_surface *surface, uint32_t
 {
   struct imv_window *window = data;
   xdg_surface_ack_configure(surface, serial);
+  window->xdg_configured = true;
+  update_scale(data);
   wl_surface_commit(window->wl_surface);
 }
 
@@ -599,8 +610,11 @@ static void toplevel_configure(void *data, struct xdg_toplevel *toplevel,
   (void)toplevel;
 
   struct imv_window *window = data;
-  window->width = width;
-  window->height = height;
+
+  if (width > 0 && height > 0) {
+    window->width = width;
+    window->height = height;
+  }
   window->fullscreen = false;
 
   enum xdg_toplevel_state *state;
@@ -618,10 +632,10 @@ static void toplevel_configure(void *data, struct xdg_toplevel *toplevel,
     .type = IMV_EVENT_RESIZE,
     .data = {
       .resize = {
-        .width = width,
-        .height = height,
-        .buffer_width = width * window->scale,
-        .buffer_height = height * window->scale
+        .width = window->width,
+        .height = window->height,
+        .buffer_width = window->width * window->scale,
+        .buffer_height = window->height * window->scale
       }
     }
   };
@@ -657,7 +671,7 @@ static void connect_to_wayland(struct imv_window *window)
   assert(window->wl_registry);
 
   wl_registry_add_listener(window->wl_registry, &registry_listener, window);
-  wl_display_dispatch(window->wl_display);
+  wl_display_roundtrip(window->wl_display);
   assert(window->wl_compositor);
   assert(window->wl_xdg);
   assert(window->wl_seat);
@@ -702,6 +716,9 @@ static void create_window(struct imv_window *window, int width, int height,
   window->egl_surface = eglCreateWindowSurface(window->egl_display, config, window->egl_window, NULL);
   eglMakeCurrent(window->egl_display, window->egl_surface, window->egl_surface, window->egl_context);
 
+  window->width = width;
+  window->height = height;
+
   wl_surface_commit(window->wl_surface);
   wl_display_roundtrip(window->wl_display);
 }
@@ -852,7 +869,9 @@ void imv_window_get_mouse_position(struct imv_window *window, double *x, double
 
 void imv_window_present(struct imv_window *window)
 {
-  eglSwapBuffers(window->egl_display, window->egl_surface);
+  if (window->xdg_configured) {
+    eglSwapBuffers(window->egl_display, window->egl_surface);
+  }
 }
 
 void imv_window_wait_for_event(struct imv_window *window, double timeout)
